# Version 0.1
#
#  PhylogeneticTree is copyright 2005 Jeffrey G. Pittman.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#  MA 02111-1307 USA

import csv, string, cStringIO

import os, sys, math, random

from reportlab.lib import colors
from reportlab.pdfgen.canvas import Canvas

from reportlab.graphics.shapes import Group
from reportlab.graphics.shapes import Drawing

import csv
import calendar
# the dateutil functionality should perhaps be added:
#from dateutil.relativedate import *
#from dateutil.rrule import *
#from dateutil.parser import *
from datetime import date
from datetime import datetime
from datetime import timedelta

from Products.PhylogeneticTree.utils import getrlcolor
from Products.PhylogeneticTree.utils import colorstr
from Products.PhylogeneticTree.utils import wordWrap
from Products.PhylogeneticTree.utils import rgb256

from Products.PhylogeneticTree.graphics import Line
from Products.PhylogeneticTree.graphics import DashedLine
from Products.PhylogeneticTree.graphics import Circle
from Products.PhylogeneticTree.graphics import Rectangle
from Products.PhylogeneticTree.graphics import IDRectangle
from Products.PhylogeneticTree.graphics import Text
from Products.PhylogeneticTree.graphics import TransformedText
from Products.PhylogeneticTree.graphics import AlignedText
from Products.PhylogeneticTree.graphics import TransformedAndAlignedText

from reportlab.graphics.shapes import Line as RLLine
from reportlab.graphics.shapes import Rect as RLRect
from reportlab.graphics.shapes import Circle as RLCircle
from reportlab.graphics.shapes import Ellipse as RLEllipse
from reportlab.graphics.shapes import PolyLine as RLPolyLine
from reportlab.graphics.charts.textlabels import Label

class Character:
    def __init__(self, number, short_description, long_description):
        self.number = number
        self.short_description = short_description
        self.long_description = long_description

class Node:
    def __init__(self, parent_name, taxon_name, characters, 
                 row_position, column_position, font_size,
                 line_style, line_base_style, line_stroke_width, line_direction, branch_color,
                 node_style, node_size, node_skew, node_color):
        self.parent_name = parent_name
        self.taxon_name = taxon_name
        self.characters = characters
        self.row_position = row_position
        self.column_position = column_position
        self.font_size = font_size
        self.line_style = line_style
        self.line_base_style = line_base_style
        self.line_stroke_width = line_stroke_width
        self.line_direction = line_direction
        self.branch_color = branch_color
        self.node_style = node_style
        self.node_size = node_size
        self.node_skew = node_skew
        self.node_color = node_color

        self.descendants = []

        self.location = None

    def setDescendant(self, descendant):
        self.descendants.append(descendant)

    def hasSingleRightDescendant(self):
        right_count = 0
        for descendant in self.descendants:
            if descendant.line_direction == 'right':
                right_count += 1
        if right_count == 1:
            return 1
        return 0

    def getRightDescendants(self):
        right_descendants = []
        for descendant in self.descendants:
            if descendant.line_direction == 'right':
                right_descendants.append(descendant)
        return right_descendants

    def setLocation(self, location):
        self.location = location

class PhylogeneticTreeDrawingEngine:
    """
    The basic structure of SVG emitting methods is a modification of:
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/325823,
    whereby the strarray() method emits a list of lines that are
    added to a growing SVG file.

    SVG help at www.carto.net was also very useful.    
    """
    def __init__(self, 
                 left_margin, right_margin, top_margin, bottom_margin, 
                 rows, row_height, column_width, 
                 characters, nodes):
        self.left_margin = left_margin
        self.right_margin = right_margin
        self.top_margin = top_margin
        self.bottom_margin = bottom_margin
        self.rows = rows
        self.row_height = row_height
        self.column_width = column_width
        self.characters = characters
        self.nodes = nodes

        self.parents_and_descendants = {}
        for node in self.nodes:
            if not self.parents_and_descendants.has_key(node.parent_name):
                self.parents_and_descendants[node.parent_name] = []
                self.parents_and_descendants[node.parent_name].append(node)
            else:
                self.parents_and_descendants[node.parent_name].append(node)

        for parent_name in self.parents_and_descendants.keys():
            if parent_name != 'root':
                parent = self.getNode(parent_name)
                for descendant in self.parents_and_descendants[parent_name]:
                    parent.setDescendant(descendant)

        self.total_width = 0
        self.total_height = 0

        # rightmost_descendant is used for a stop, in processing
        rightmost_descendant = None
      
        # make a preliminary pass to determine item_width;
        # we only go right here (for the "baseline"):
        item_width = 0
        node = self.nodes[0]
        while 1:
            right_descendants = node.getRightDescendants()
            if len(right_descendants) == 1:
                item_width += self.column_width
                node = right_descendants[0]
                right_most_descendant = node
            if len(right_descendants) == 0:
                break

        # the root is the first entry in the list
        self.root = self.nodes[0]

        # if the skew of the first left descendant is non-zero, then add
        # the amount of the skew to the width
        #left_offset_for_skew = 0
        #if self.root.descendants[0].node_skew < 0:
            #first_left = self.root.descendants[0]
            #left_offset_for_skew = int(math.fabs((first_left.row_position * float(self.row_height)) * math.tan(math.radians(first_left.node_skew))))
            #item_width += left_offset_for_skew

        # for the simple hard-coded entry system, just set the locations
        # using the values entered for the tree
        max_row_position = 0.0
        max_column_position = 0.0
        for node in self.nodes:
            node.setLocation((int(node.column_position * self.column_width), int(node.row_position * self.row_height)))
            if node.row_position > max_row_position:
                max_row_position = node.row_position
            if node.column_position > max_column_position:
                max_column_position = node.column_position

        # be careful to take into account left_offset_for_skew
        #self.root.setLocation((left_offset_for_skew, 0))

        # set the other node locations
        #for node in self.root.descendants:
            #self.traverseAndSetLocation(node)

        panel_width = item_width

        # try hard-coded method
        panel_width = max_column_position * self.column_width
        self.total_width = self.left_margin + panel_width + self.right_margin

        #panel_height = (self.end_hour - self.start_hour) * self.hour_height
        panel_height = self.rows * self.row_height

        # try hard-coded method
        panel_height = max_row_position * self.row_height
        self.total_height = self.bottom_margin + panel_height + self.top_margin

    def getNode(self, taxon_name):
        for node in self.nodes:
            if node.taxon_name == taxon_name:
                return node
        return None
            
    def yUp(self, y):
        return self.total_height - float(y)

    def getHTMLForContainingImageMap(self, name):        
        var = ["<body>\n",
               "<map name=\"drawing.map\">\n"]

        # imagemap coordinates are like SVG coordinates, from the upper
        # left corder of the image, so use yUp as with SVG

        for node in self.nodes:
            if len(node.characters) > 0:
                characters_string = []
                for character_index in node.characters:
                    characters_string += "%s<br />" % self.characters[character_index-1].short_description
                characters_string = ''.join(characters_string)

                if node.node_style.startswith('square'):
                    x1 = self.left_margin + node.location[0] - (node.node_size / 2.0)
                    y1 = self.yYp(self.bottom_margin + node.location[1] + (node.node_size / 2.0)) 
                    x2 = self.left_margin + node.location[0] + (node.node_size / 2.0)
                    y2 = self.yYp(self.bottom_margin + node.location[1] + (node.node_size / 2.0)) 
                    var += "<area onmouseover=\"return escape(\'%s\')\" shape=\"rect\" coords=\"%d, %d, %d, %d\">\n" % (characters_string, x1, y1, x2, y2)
                else:
                    x1 = self.left_margin + node.location[0] + (node.node_size / 2.0)
                    y1 = self.yUp(self.bottom_margin + node.location[1] + (node.node_size / 2.0))
                    radius = node.node_size
                    var += "<area onmouseover=\"return escape(\'%s\')\" shape=\"circle\" coords=\"%d, %d, %d\">\n" % (characters_string, x1, y1, radius) 

        var += ["</map>\n",
                "<img usemap=\"drawing.map\" src=\"drawing.png\">\n",
                "</body>\n"]

        return var

    def getHTMLForContainingSVG(self, name):        
        var = ["<body>\n",
               "<H1>Instructions below don't apply to all browsers.</H1>\n",
               "<ul>\n",
               "    <li><b>Zooming</b>: use the CTRL-Key (or Apple-Key) to drag a rectangle and zoom in.\n",
               "        <br />Use the right mouse button (CTRL-Key plus mouse click on the Mac) to get a context menu, <br />where you can choose step-by-step zooming and can go back to the original extent.</li>\n",
               "    <li><b>Zoom Out</b>: like above, but with SHIFT-Key pressed.</li>\n",
               "    <li><b>Panning</b>: Press the ALT-Key and move the mouse cursor while a hand-icon appears.</li>\n",
               "    <li><b>Anti-Aliasing</b>: May be turned on and off with the help of the context menu. (Higher Quality).</li>\n",
               "</ul>\n",
               "<h2>SVG Phylogenetic Tree</h2>\n",
               "          <embed width=\"100%%\" height=\"660px\" src=\"%s.svg\" type=\"image/svg+xml\" />\n" % name,
               "</body>\n"]
               #"              <b>For viewing SVG here, perhaps a browser update would help.<br />Perhaps not, as SVG browser support is now incomplete.</b>",
               #"          <object width=\"600\" height=\"660\" data=\"%s.svg\" type=\"image/svg+xml\" >\n" % name,
               #"          <iframe width=\"600\" height=\"660\" src=\"%s.svg\" frameborder=\"0\" marginwidth=\"0\" marginheight=\"0\" >\n" % name,
               #"              <embed width=\"600\" height=\"660\" src=\"%s.svg\" type=\"image/svg+xml\" />\n" % name,
            #["  <script type=\"text/javascript\">\n",
            #["<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">\n",
             #"<html>\n",
             #"<head>\n",
             #"  <meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\">\n",
             #"  <meta http-equiv=\"imagetoolbar\" content=\"no\">\n",
             #"  <meta name=\"keywords\" content=\"geologic time scale svg scalable vector graphics\">\n",
             #"  <meta name=\"description\" content=\"SVG Phylogenetic Tree\">\n",
             #"  <style type=\"text/css\">\n",
             #"      body {background-color:#ffef95;font-family:Arial,Helvetica,Geneva,sans-serif;color:#B42D25;}\n",
             #"      .bigger {font-weight:bold;}\n",
             #"      .smaller {font-size:small;}\n",
             #"  </style>\n",
             #"      function hilite_elem(checkbox, element_name){\n",
             #"        // For each element, get the element's style object, then set\n",
             #"        // its visibility according to the state of the checkbox.\n",
             #"        var svgobj = document.embeds['phylogenetictree'].getSVGDocument().getElementById(element_name);\n",
             #"        if (!checkbox.checked){\n",
             #"            // Hide layer.\n",
             #"            svgobj.setAttributeNS(null,'visibility','hidden');\n",
             #"        } else {\n",
             #"            // Show layer.\n",
             #"            svgobj.setAttributeNS(null,'visibility','visible');\n",
             #"        }\n",
             #"      }\n",
             #"  </script>\n",
             #"  <title>SVG Phylogenetic Tree</title>\n",
             #"</head>\n",
             #"          <object width=\"600\" height=\"660\" data=\"%s.svg\" name=\"phylogenetictree\" type=\"image/svg+xml\"/>\n" % name,
             #"            <br>\n",
             #"            <form name=\"hilite_form\" action=\"\">\n",
             #"              <table border=\"0\" cellpadding=\"0\" cellspacing=\"2\">\n",
             #"                <tr>\n",
             #"                  <td colspan=\"2\">Visible Layers\n",
             #"                  </td>\n",
             #"                </tr>\n",
             #"                <tr>\n",
             #"                  <td><input type=\"checkbox\" value=\"\" onclick=\"hilite_elem(this,'bar_scale')\">&nbsp;Bar Scale</TD>\n",
             #"                  <td><input type=\"checkbox\" value=\"\" onclick=\"hilite_elem(this,'header_labels')\">&nbsp;Header Labels</TD>\n",
             #"                </tr>\n",
             #"               </table>\n",
             #"               <script type=\"text/javascript\">\n",
             #"                 // Make sure all checkboxes are checked whenever the page is reloaded in the browser.\n",
             #"                 for (var i = 0;  i < document.hilite_form.elements.length;  i++)\n",
             #"                   if (document.hilite_form.elements[i].type == 'checkbox'){\n",
             #"                     document.hilite_form.elements[i].checked = true;\n",
             #"                   }  \n",
             #"               </script>\n",
             #"            </form>\n",
             #"      <table border=\"0\" cellpadding=\"8\" cellspacing=\"0\">\n",
             #"        <tr>\n",
             #"          <td align=\"center\" valign=\"top\">\n",
             #"          </td>\n",
             #"          <td valign=\"top\">\n",
             #"            <p>Use mouse functions to change display:</p>\n",
             #"            <ul>\n",
             #"              <li><span class=\"bigger\">Layers (if enabled)</span>: Use the checkboxes at bottom to toggle items on and off.</li>\n",
             #"            </ul>\n",
             #"            <p>Use keyboard functions to change display:</p>\n",
             #"            <ul>\n",
             #"              <li><span class=\"bigger\">Zooming</span>: use the CTRL-Key (or Apple-Key) to drag a rectangle and zoom in.\n",
             #"                  <br>Use the right mouse button (CTRL-Key&nbsp;+ mouse click on the Mac) to get a context menu, where you can choose step-by-step zooming and can go back to the original extent.</li>\n",
             #"              <li><span class=\"bigger\">Zoom Out</span>: like above, but with SHIFT-Key pressed.</li>\n",
             #"              <li><span class=\"bigger\">Panning</span>: Press the ALT-Key and move the mouse cursor while a \"hand-icon\" appears.</li>\n",
             #"              <li><span class=\"bigger\">Anti-Aliasing</span>: May be turned on and off with the help of the context menu. (&quot;Higher Quality&quot;).</li>\n",
             #"            </ul>\n",
             #"          </td>\n",
             #"        </tr>\n",
             #"      </table>\n",
             #"</body></html>\n"]
        return var

    def getHTML(self):
        var = ["<body>\n"]

        var += "<H1>%s</H1>\n" % ('Characters')

        for character in self.characters:
            var += "<dl>\n"
            var += "    <dt>%d: %s</dt>\n" % (character.number, character.short_description)
            var += "    <dd>%s</dd>\n" % (character.long_description)
            var += "</dl>\n"

        var += "<H1>%s</H1>\n" % ('Taxa')

        for node in self.nodes:
            if not node.taxon_name.startswith('xxx'):
                var += "<dl>\n"
                var += "    <dt>%s</dt>\n" % (node.taxon_name)
                for character_index in node.characters:
                    character = self.characters[character_index-1]
                    var += "    <dd>%d: %s</dd>\n" % (character.number, character.long_description)
                var += "</dl>\n"

        var += "</body>\n"

        return var

    def getCSS(self):
        css = \
          ["\n",
           "        <style type=\"text/css\">\n",
           "            <![CDATA[\n",
           "                body {font-family:Arial,Helvetica,Geneva,sans-serif;}\n",
           "                .str0 {\n",
           "                    stroke:#000000;\n",
           "                    stroke-width:0.1\n",
           "                }\n",
           "            ]]>\n",
           "        </style>\n",
           "\n"]
        return css

    def getEcmascript(self):
        """
        Note: at one time there were three description lines in the mouseover.
        These are commented out. This could be useful to have more info
            in the tooltip.
        """
        ecmascript = \
            ["        <script language=\"Javascript\"><![CDATA[\n",
             "        // This script from http://svg-whiz.com/svg/Tooltip.svg\n",
             "        var SVGDocument = null;\n",
             "        var SVGRoot = null;\n",
             "        var SVGViewBox = null;\n",
             "        var svgns = 'http://www.w3.org/2000/svg';\n",
             "        var xlinkns = 'http://www.w3.org/1999/xlink';\n",
             "        var toolTip = null;\n",
             "        var TrueCoords = null;\n",
             "        var tipBox = null;\n",
             "        var tipText = null;\n",
             "        var tipTitle = null;\n",
             #"        var tipDesc1 = null;\n",
             #"        var tipDesc2 = null;\n",
             #"        var tipDesc3 = null;\n",
             "\n",
             "        var lastElement = null;\n",
             "        var titleText = '';\n",
             #"        var titleDesc1 = '';\n",
             #"        var titleDesc2 = '';\n",
             #"        var titleDesc3 = '';\n",
             "        function Init(evt)\n",
             "        {\n",
             "        SVGDocument = evt.target.ownerDocument;\n",
             "        SVGRoot = document.documentElement;\n",
             "        TrueCoords = SVGRoot.createSVGPoint();\n",
             "        \n",
             "        toolTip = SVGDocument.getElementById('ToolTip');\n",
             "        tipBox = SVGDocument.getElementById('TipBox');\n",
             "        tipText = SVGDocument.getElementById('TipText');\n",
             "        tipTitle = SVGDocument.getElementById('TipTitle');\n",
             #"        tipDesc1 = SVGDocument.getElementById('tipDesc1');\n",
             #"        tipDesc2 = SVGDocument.getElementById('tipDesc2');\n",
             #"        tipDesc3 = SVGDocument.getElementById('tipDesc3');\n",
             "        };\n",
             "\n",
             "        function GetTrueCoords(evt)\n",
             "        {\n",
             "           // find the current zoom level and pan setting, and adjust the reported\n",
             "           //    mouse position accordingly\n",
             "           var newScale = SVGRoot.currentScale;\n",
             "           var translation = SVGRoot.currentTranslate;\n",
             "           TrueCoords.x = (evt.clientX - translation.x)/newScale;\n",
             "           TrueCoords.y = (evt.clientY - translation.y)/newScale;\n",
             "        };\n",
             "\n",
             "        function ShowTooltip(evt, turnOn)\n",
             "        {\n",
             "           try\n",
             "           {\n",
             "              if (!evt || !turnOn)\n",
             "              {\n",
             "                 toolTip.setAttributeNS(null, 'display', 'none');\n",
             "              }\n",
             "              else\n",
             "              {\n",
             "                 var tipScale = 1/SVGRoot.currentScale;\n",
             "                 var textWidth = 0;\n",
             "                 var tspanWidth = 0;\n",
             "                 var boxHeight = 20;\n",
             "\n",
             "                 tipBox.setAttributeNS(null, 'transform', 'scale(' + tipScale + ',' + tipScale + ')' );\n",
             "                 tipText.setAttributeNS(null, 'transform', 'scale(' + tipScale + ',' + tipScale + ')' );\n",
             "\n",
             "                 var targetElement = evt.target;\n",
             "\n",
             "                 if ( lastElement != targetElement )\n",
             "                 {\n",
             "                    var targetTitle = targetElement.getElementsByTagName('title').item(0);\n",
             "                    if ( targetTitle )\n",
             "                    {\n",
             "                       titleText = targetTitle.firstChild.nodeValue;\n",
             "                       tipTitle.firstChild.nodeValue = titleText;\n",
             "                    }\n",
             "\n",
             #"                    var targetDesc1 = targetElement.getElementsByTagName('desc1').item(0);\n",
             #"                    if ( targetDesc1 )\n",
             #"                    {\n",
             #"                       titleDesc1 = targetDesc1.firstChild.nodeValue;\n",
             #"                       tipDesc1.firstChild.nodeValue = titleDesc1;\n",
             #"                    }\n",
             #"\n",
             #"                    var targetDesc2 = targetElement.getElementsByTagName('desc2').item(0);\n",
             #"                    if ( targetDesc2 )\n",
             #"                    {\n",
             #"                       titleDesc2 = targetDesc2.firstChild.nodeValue;\n",
             #"                       tipDesc2.firstChild.nodeValue = titleDesc2;\n",
             #"                    }\n",
             #"\n",
             #"                    var targetDesc3 = targetElement.getElementsByTagName('desc3').item(0);\n",
             #"                    if ( targetDesc3 )\n",
             #"                    {\n",
             #"                       titleDesc3 = targetDesc3.firstChild.nodeValue;\n",
             #"                       tipDesc3.firstChild.nodeValue = titleDesc3;\n",
             #"                    }\n",
             "                 }\n",
             "\n",
             "                 var xPos = TrueCoords.x + (10 * tipScale);\n",
             "                 var yPos = TrueCoords.y + (10 * tipScale);\n",
             "\n",
             "                 //return rectangle around object as SVGRect object\n",
             "                 var outline = tipText.getBBox();\n",
             "                 tipBox.setAttributeNS(null, 'width', Number(outline.width) + 10);\n",
             "                 tipBox.setAttributeNS(null, 'height', Number(outline.height) + 10);\n",
             "\n",
             "                 toolTip.setAttributeNS(null, 'transform', 'translate(' + xPos + ',' + yPos + ')');\n",
             "                 toolTip.setAttributeNS(null, 'display', 'inline');\n",
             "              }\n",
             "           }\n",
             "           catch(er){}\n",
             "         };\n",
             "        ]]></script>\n"]
        return ecmascript

    def getOpenerSection(self):
        opener_section = \
          ["<?xml version=\"1.0\"?>\n",
           "<svg  id=\"phylogenetictree\" xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" height=\"%d\" width=\"%d\">\n" % (self.total_height,self.total_width)]
           #"<svg  id=\"phylogenetictree\" xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" onload=\"Init(evt)\" onmousemove=\"GetTrueCoords(evt); ShowTooltip(evt, true)\" onmouseout=\"ShowTooltip(evt, false)\" height=\"%d\" width=\"%d\">\n" % (self.total_height,self.total_width)]
        return opener_section
                      
    def getDefsSection(self):
        defs_section = ["    <defs>\n"]
        defs_section += self.getCSS()
        defs_section += self.getEcmascript()
        defs_section += ["    </defs>\n"]
        return defs_section
                    
    def getBottomPopupSection(self):
        bottom_svg = \
          ["    <g id='ToolTip' opacity='0.8' display='none' pointer-events='none'>\n",
           "        <rect id='TipBox' x='0' y='5' width='88' height='20' rx='2' ry='2' fill='white' stroke='black'/>\n",
           "        <text id='TipText' x='5' y='20' font-family='Arial' font-size='12'>\n",
           "            <tspan id='TipTitle' x='5' font-weight='bold'> </tspan>\n",
           #"         <tspan id='tipDesc1' x='5' dy='1.2em' fill='blue'> </tspan>\n",
           #"         <tspan id='tipDesc2' x='5' dy='1.2em' fill='blue'> </tspan>\n",
           #"         <tspan id='tipDesc3' x='5' dy='1.2em' fill='blue'> </tspan>\n",
           "        </text>\n",
           "    </g>\n"]
        return bottom_svg
    
    def getCloserSection(self):
        closer_section = ["</svg>\n"]
        return closer_section

    def traverseAndSetLocation(self, node):
        parent = self.getNode(node.parent_name)
        y_difference = node.row_position - parent.row_position
        if node.line_direction == 'left':
            if node.node_skew < 0 or node.node_skew > 0:
                x_offset = int((y_difference * float(self.row_height)) * math.tan(math.radians(node.node_skew)))
                x = parent.location[0] + x_offset
            else:
                x = parent.location[0]
            y = parent.location[1] + int(y_difference * float(self.row_height))
            node.setLocation((x, y))
        elif node.line_direction == 'right':
            x = parent.location[0] + int(node.column_position * float(self.column_width))
            y = parent.location[1] + int(y_difference * float(self.row_height))
            node.setLocation((x, y))
        elif node.line_direction == 'up':
            x = parent.location[0]
            y = parent.location[1] + int(y_difference * float(self.row_height))
            node.setLocation((x, y))
            
        for descendant in node.descendants:
            self.traverseAndSetLocation(descendant)
        
    def traverseAndDrawBranches(self, var, node):
        parent = self.getNode(node.parent_name)
        line = Line((parent.location[0], 
                     self.yUp(parent.location[1])),
                     (node.location[0], 
                     self.yUp(node.location[1])),
                     node.line_stroke_width,
                     rgb256(node.branch_color.rgb()))
        
        var += line.strarray()
            
        for descendant in node.descendants:
            self.traverseAndDrawBranches(var, descendant)
        
    def getPhylogeneticTreeSection(self):
        # For SVG, the 0,0 point is in the upper left corner, as compared
        # to the lower left corner for the reportlab toolkit. We handle
        # this difference in the y-axis by use of the yUp() method.

        var = []
    
        # draw the branches
        for node in self.root.descendants:
            self.traverseAndDrawBranches(var, node)

        # draw the nodes
        for node in self.nodes:
            if not node.taxon_name.startswith('xxx'):
                if node.node_style.startswith('circle'):
                    circle = Circle((node.location[0],
                                     self.yUp(node.location[1])),
                                    node.node_size,
                                    rgb256(node.node_color.rgb()))
                    var += circle.strarray()
                elif node.node_style.startswith('square'):
                    rectangle = Rectangle((node.location[0] - (node.node_size / 2.0), 
                                           self.yUp(node.location[1] - (node.node_size / 2.0))),
                                          node.node_size,
                                          node.node_size,
                                          0.1,
                                          rgb256(node.node_color.rgb()),
                                          rgb256(node.node_color.rgb()))
                    var += rectangle.strarray()
                    
        # draw the labels
        for node in self.nodes:
            if not node.taxon_name.startswith('xxx'):
                if len(node.descendants) == 0: # leaf node
                    label = TransformedAndAlignedText((node.location[0] + 8, 
                                                       self.yUp(node.location[1] + 8)),
                                                      "%s" % node.taxon_name,
                                                      rgb256(node.node_color.rgb()),
                                                      node.font_size,
                                                      -45,
                                                      'start')
                    var += label.strarray()
                else: # interior node
                    label = AlignedText((node.location[0] + 8, 
                                         self.yUp(node.location[1] - 8)),
                                        "%s" % node.taxon_name,
                                        rgb256(node.node_color.rgb()),
                                        node.font_size,
                                        "start")
                    var += label.strarray()
        
        return var

    def getSVG(self):
        svg_code = self.getOpenerSection()
        #svg_code += self.getDefsSection()
        svg_code += self.getPhylogeneticTreeSection()
        #svg_code += self.getBottomPopupSection()
        svg_code += self.getCloserSection()
        return svg_code

    def traverseAndDrawBranchesRL(self, g, node):
        parent = self.getNode(node.parent_name)
        g.add(RLLine(parent.location[0], parent.location[1],
                     node.location[0], node.location[1],
                     strokeWidth = node.line_stroke_width,
                     strokeColor = node.branch_color))
            
        for descendant in node.descendants:
            self.traverseAndDrawBranchesRL(g, descendant)
        
    def getDrawing(self):
        '''Phylogenetic Tree Plot'''

        D = Drawing(self.total_width, self.total_height) 
        g = Group(transform=(1,0,0,1,0,0))
    
        g.translate(self.left_margin, self.bottom_margin)

        # draw the branches
        for node in self.root.descendants:
            self.traverseAndDrawBranchesRL(g, node)

        # draw the nodes
        for node in self.nodes:
            if not node.taxon_name.startswith('xxx'):
                if node.node_style.startswith('circle'):
                    g.add(RLCircle(node.location[0], 
                                   node.location[1],
                                   node.node_size,
                                   fillColor=node.node_color,
                                   strokeColor=node.node_color,
                                   strokeWidth=0.1))
                elif node.node_style.startswith('square'):
                    g.add(RLRect(node.location[0] - (node.node_size / 2.0),
                                 node.location[1] - (node.node_size / 2.0),
                                 node.node_size,
                                 node.node_size,
                                 fillColor=node.node_color,
                                 strokeColor=node.node_color,
                                 strokeWidth=0.5))
                    
        # draw the labels
        for node in self.nodes:
            if not node.taxon_name.startswith('xxx'):
                if len(node.descendants) == 0: # leaf node
                    label = Label()
                    label.setOrigin(node.location[0] + 8,
                                    node.location[1] + 8)
                    label.angle = 45
                    label.setText("%s" % node.taxon_name)
                    label.textAnchor = 'start'
                    label.boxAnchor = 'w'
                    label.fontName = 'Helvetica'
                    label.fontSize = node.font_size
                    g.add(label)
                else: # interior node
                    label = Label()
                    label.setOrigin(node.location[0]+8, 
                                    node.location[1]-8)
                    label.angle = 0
                    label.setText("%s" % node.taxon_name)
                    label.textAnchor = 'start'
                    label.boxAnchor = 'w'
                    label.fontName = 'Helvetica'
                    label.fontSize = node.font_size
                    g.add(label)
        
        D.add(g)
        g = Group(transform=(1,0,0,1,0,0))
        return D

class PhylogeneticTreeDrawing:
    def __init__(self, render_types, 
                 left_margin, right_margin, 
                 top_margin, bottom_margin, 
                 scale_to_fit, 
                 rows, row_height, 
                 column_width, 
                 characters_input,
                 nodes_input):
        self.render_types = render_types
        self.left_margin = left_margin
        self.right_margin = right_margin
        self.top_margin = top_margin
        self.bottom_margin = bottom_margin
        self.scale_to_fit = scale_to_fit
        self.rows = rows
        self.row_height = row_height
        self.column_width = column_width
        self.characters_input = characters_input
        self.nodes_input = nodes_input

        name = "phylogenetictree"

        characters = []
        for character_string in self.characters_input:
            number, short_description, long_description = eval('[%s]' % character_string)
            character = Character(number, short_description, long_description)
            characters.append(character)

        nodes = []
        for node_string in self.nodes_input:
            parent_name, taxon_name, characters_string, row_position, column_position, font_size, line_style, line_base_style, line_stroke_width, line_direction, line_red, line_green, line_blue, node_style, node_size, node_skew, node_red, node_green, node_blue = eval('[%s]' % node_string)
            
            # characters_string has already been validated, so parse and
            # set and array of ints
            characters_string = characters_string[1:-1]
            node_characters = []
            if len(characters_string) > 0:
                node_characters = [int(c) for c in characters_string.split(',')]

            node = Node(parent_name, taxon_name, node_characters, 
                        row_position, column_position, font_size,
                        line_style, line_base_style, line_stroke_width, line_direction,
                        getrlcolor(line_red, line_green, line_blue), 
                        node_style, node_size, node_skew,
                        getrlcolor(node_red, node_green, node_blue))

            nodes.append(node)

        self.phylogenetic_tree = PhylogeneticTreeDrawingEngine(self.left_margin, 
                                                               self.right_margin, 
                                                               self.top_margin, 
                                                               self.bottom_margin, 
                                                               self.rows,
                                                               self.row_height,
                                                               self.column_width,
                                                               characters,
                                                               nodes)

    def getDrawing(self):
        self.drawing = self.phylogenetic_tree.getDrawing()

        if self.scale_to_fit:
            scale_factor = float(self.height) / self.phylogenetic_tree.total_height
            self.drawing.scale(scale_factor,scale_factor)

        return self.drawing

    def getSVG(self):
        return self.phylogenetic_tree.getSVG()

    def getHTML(self):
        return self.phylogenetic_tree.getHTML()

    def getHTMLForContainingSVG(self, name):
        return self.phylogenetic_tree.getHTMLForContainingSVG(name)

    def getHTMLForContainingImageMap(self, name):
        return self.phylogenetic_tree.getHTMLForContainingImageMap(name)


